/*****************************************************************************
Copyright (c) 2011 SmartVault, Inc.

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 ******************************************************************************/
package com.smartvault.net;

import java.util.UUID;
import java.io.*;
import java.nio.*;
import java.nio.charset.*;

public class MultipartMimeWriter
{
	private static final String APPLICATION_X_PROTOBUF = "application/x-protobuf"; //$NON-NLS-1$
	private static final String CONTENT_LENGTH = "Content-Length"; //$NON-NLS-1$
	private static final String CONTENT_TYPE = "Content-Type"; //$NON-NLS-1$
	private static final String BOUNDARY_BYTES_MUST_BE_7BIT_ENCODED = "Boundary Bytes must be 7bit encoded"; //$NON-NLS-1$
	private static final String BOUNDARY_IS_TOO_LONG = "Boundary is too long"; //$NON-NLS-1$
	private static final String US_ASCII = "US-ASCII"; //$NON-NLS-1$
	private OutputStream writer;
	private byte[] boundary;
	private static byte[] MimeBoundaryStartBinary = new byte[] {(byte)'-', (byte)'-'};
	private static byte[] CRLFBinary = new byte[] { (byte)'\r', (byte)'\n' };
	private static final String CRLF = "\r\n"; //$NON-NLS-1$
	
    public static byte[] getByteArray(UUID uuid) throws Exception
    {
		Charset charset = Charset.forName(US_ASCII);
		CharsetEncoder encoder = charset.newEncoder();
		ByteBuffer bb = null;
		
		CharBuffer cb = CharBuffer.wrap(uuid.toString());
		bb = encoder.encode(cb);
		return bb.array();
    }
	
	public MultipartMimeWriter(OutputStream s) throws Exception
	{	
		this(s, getByteArray(UUID.randomUUID()));
	}

	private MultipartMimeWriter(OutputStream s, byte[] boundary)  throws IllegalArgumentException
	{
		this.writer = s;

		// length of 1 .. 71
		if (boundary.length < 1 || boundary.length > 71)
		{
			throw new IllegalArgumentException(BOUNDARY_IS_TOO_LONG);
		}

		// Boundary must be 7bit encoded
		for (int i = 0; i < boundary.length; ++i)
		{
			if (boundary[i] > 127)
			{
				throw new IllegalArgumentException(BOUNDARY_BYTES_MUST_BE_7BIT_ENCODED);
			}
		}

		this.boundary = boundary;
	}


	private void writeHeader(String name, String value) throws Exception
	{
		String tmp = name + ": " + value + CRLF; //$NON-NLS-1$
		
		Charset charset = Charset.forName(US_ASCII);
		CharsetEncoder encoder = charset.newEncoder();
		ByteBuffer bb = null;
		
		CharBuffer cb = CharBuffer.wrap(tmp);
		bb = encoder.encode(cb);
		byte[] header = bb.array();
		this.writer.write(header, 0, header.length);
	}

	private void writeContentType(String contentType) throws Exception
	{
		writeHeader(CONTENT_TYPE, contentType);
	}

	private void writeBoundary(boolean last) throws Exception
	{
		// write boundary
		this.writer.write(MultipartMimeWriter.MimeBoundaryStartBinary, 0, MultipartMimeWriter.MimeBoundaryStartBinary.length);
		this.writer.write(this.boundary, 0, this.boundary.length);
		if (last)
		{
			this.writer.write(MultipartMimeWriter.MimeBoundaryStartBinary, 0, MultipartMimeWriter.MimeBoundaryStartBinary.length);
		}
		this.writer.write(MultipartMimeWriter.CRLFBinary, 0, MultipartMimeWriter.CRLFBinary.length);
	}
	
	public void write(com.google.protobuf.GeneratedMessage payload) throws Exception
	{	
		// boundary
		writeBoundary(false);      

		// content type
		writeContentType(APPLICATION_X_PROTOBUF);

		// end header
		this.writer.write(MultipartMimeWriter.CRLFBinary, 0, MultipartMimeWriter.CRLFBinary.length);
		
		payload.writeTo(this.writer);			

		// carriage return
		this.writer.write(MultipartMimeWriter.CRLFBinary, 0, MultipartMimeWriter.CRLFBinary.length);
	}
	
	public void write(InputStream stream) throws Exception
	{		

		// boundary
		writeBoundary(false);  

		// end header
		this.writer.write(MultipartMimeWriter.CRLFBinary, 0, MultipartMimeWriter.CRLFBinary.length);

		// write body
		byte[] buffer = new byte[16000];	
		
		int read = 16000;
		long total = 0;
		
		int len;
		while ((len = stream.read(buffer, 0, read)) > 0)
		{
		
			//debug("reading");
			//debug(Integer.toString(len));
			this.writer.write(buffer, 0, len);
			total += len;
		}
		
		// carriage return
		this.writer.write(MultipartMimeWriter.CRLFBinary, 0, MultipartMimeWriter.CRLFBinary.length);
	}

	public void complete() throws Exception
	{
		writeBoundary(true);
	}
}