/*****************************************************************************
Copyright (c) 2011 SmartVault, Inc.

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 ******************************************************************************/
package com.smartvault.rest;

import com.smartvault.net.*;
import com.smartvault.exceptions.*;
import java.io.InputStream;

import com.smartvault.proto.CopyNodeProto.CopyNodeRequest;
import com.smartvault.proto.CreateFolderProto.CreateFolderRequest;
import com.smartvault.proto.NodeProto.NodeResponse;
import com.smartvault.proto.CreateAppContextProto.CreateAppContextRequest;
import com.smartvault.proto.ReplaceOptionsProto.ReplaceOptions;
import com.smartvault.proto.UpdateNodeProto;
import com.smartvault.proto.UpdateNodeProto.ChangeNodeRequest;
import com.smartvault.proto.UpdateNodeProto.MoveNodeRequest;
import com.smartvault.proto.UploadFileProto.UploadFileRequest;
import com.smartvault.proto.CreateAppAliasProto.CreateAppAliasRequest;

import java.util.*;

public class DelegatedProtocol {
	private UserCfg cfg;

	public DelegatedProtocol(UserCfg cfg) {
		this.cfg = cfg;
	}

	private ClientNetContext createClient(String noun, String verb,
			Map<String, String> m) throws Exception {
		WebRequestInfo webRequestInfo = new WebRequestInfo(
				this.cfg.getRootUri(),
				this.cfg.getUserName(),
				this.cfg.getAccessToken()
		);

		return ClientNetContext.open(webRequestInfo, noun, verb, m);
	}

	private ClientNetContext createClient(String noun, String verb)
			throws Exception {
		WebRequestInfo webRequestInfo = new WebRequestInfo(
				this.cfg.getRootUri(), this.cfg.getUserName(),
				this.cfg.getAccessToken());

		return ClientNetContext.open(webRequestInfo, noun, verb);
	}
	
	public NodeResponse navigate(String remotePath, NodeOptions options) 
		throws Exception {	
		ClientNetContext ctx = createClient(remotePath, ClientNetContext.GET, options.getOptions());

		return ctx.readPayloadNodeResponse();
	}

	public NodeResponse navigate(String remotePath, int children)
			throws Exception {
		NodeOptions options = new NodeOptions();
		options.children = children;
		
		return navigate(remotePath, options);	
	}

	public NodeResponse createFolder(String parent, String name) throws Exception {
		CreateFolderRequest.Builder req = CreateFolderRequest.newBuilder();	
		ClientNetContext ctx = createClient(parent + Constants.SLASH + name, ClientNetContext.PUT);		
		ctx.writePayload(req.build());		
		return ctx.readPayloadNodeResponse();
	}
	
	public NodeResponse postFile(String path, UploadFileRequest request,
			InputStream stream) throws Exception {
		ClientNetContext ctx = createClient(path, ClientNetContext.POST);
		ctx.writePayload(request, stream);
		return ctx.readPayloadNodeResponse();
	}

	public NodeResponse modifyFile(String path, String description) throws Exception
	{
		UpdateNodeProto.UpdateNodeRequest.Builder updateBuilder =
				UpdateNodeProto.UpdateNodeRequest.newBuilder();
		updateBuilder.setDescription(description);
		updateBuilder.setUpdateDescription(true);

		UpdateNodeProto.ChangeNodeRequest.Builder changeBuilder = UpdateNodeProto.ChangeNodeRequest.newBuilder();
		changeBuilder.setUpdate(updateBuilder.build());

		ClientNetContext ctx = createClient(path, ClientNetContext.POST);
		ctx.writePayload(changeBuilder.build());
		return ctx.readPayloadNodeResponse();
	}

	public NodeResponse modifyFolder(String path, String description) throws Exception
	{
		UpdateNodeProto.UpdateNodeRequest.Builder updateBuilder =
				UpdateNodeProto.UpdateNodeRequest.newBuilder();
		updateBuilder.setDescription(description);

		UpdateNodeProto.ChangeNodeRequest.Builder changeBuilder = UpdateNodeProto.ChangeNodeRequest.newBuilder();
		changeBuilder.setUpdate(updateBuilder.build());

		ClientNetContext ctx = createClient(path, ClientNetContext.POST);
		ctx.writePayload(changeBuilder.build());
		return ctx.readPayloadNodeResponse();
	}
	
	public NodeResponse copy(String pathFrom, String pathTo, ReplaceOptions replaceOption) throws Exception {
		ChangeNodeRequest.Builder request = ChangeNodeRequest.newBuilder();
		CopyNodeRequest.Builder builder = CopyNodeRequest.newBuilder();
		builder.setDstUri(pathTo);
		builder.setReplace(replaceOption);
		ClientNetContext ctx = createClient(pathFrom, ClientNetContext.POST);
		request.setCopy(builder.build());
		ctx.writePayload(request.build());
		return ctx.readPayloadNodeResponse();
	}
	
	public NodeResponse copy(String pathFrom, String pathTo) throws Exception {
		ReplaceOptions replaceOption = ReplaceOptions.Replace;
		return copy(pathFrom, pathTo, replaceOption);
	}
	
	public NodeResponse move(String pathFrom, String pathTo, ReplaceOptions replaceOption) throws Exception {
		ChangeNodeRequest.Builder request = ChangeNodeRequest.newBuilder();
		MoveNodeRequest.Builder builder = MoveNodeRequest.newBuilder();
		builder.setDstUri(pathTo);
		builder.setReplace(replaceOption);
		ClientNetContext ctx = createClient(pathFrom, ClientNetContext.POST);
		request.setMove(builder.build());
		ctx.writePayload(request.build());
		return ctx.readPayloadNodeResponse();
	}
	
	public NodeResponse move(String pathFrom, String pathTo) throws Exception {
		ReplaceOptions replaceOption = ReplaceOptions.Replace;
		return move(pathFrom, pathTo, replaceOption);
	}
	
	public NetAttachmentReader readFile(String path) throws Exception
    {
    	ClientNetContext ctx = createClient(path, ClientNetContext.GET);
    	return ctx.readAttachment();
    }
    
    public void deleteFile(String path) throws Exception { 
		ClientNetContext ctx = createClient(path, ClientNetContext.DELETE);
		ctx.readDeleteFileResponse();
	}   

	public NodeResponse provisionCompany(String provisionAt, String name,
			String applicationType, String applicationId) throws Exception {
		CreateAppContextRequest.Builder app = CreateAppContextRequest
				.newBuilder();
		app.setName(name);

		app.setProvisionAt(provisionAt);
		String pth = Constants.APP_PATH + Constants.SLASH + applicationType
				+ Constants.SLASH + applicationId;

		ClientNetContext ctx = createClient(pth, ClientNetContext.PUT);

		ctx.writePayload(app.build());

		return ctx.readPayloadNodeResponse();
	}	
	
	public NodeResponse addAlias(String referenceUri, String applicationType,
		String applicationId) throws Exception {
		
		CreateAppAliasRequest.Builder app = CreateAppAliasRequest
				.newBuilder();
		
		app.setReferenceUri(referenceUri);
		
		String pth = getAliasPath(applicationType, applicationId);
		
		ClientNetContext ctx = createClient(pth, ClientNetContext.PUT);
		
		ctx.writePayload(app.build());
		
		return ctx.readPayloadNodeResponse();
	}

	public String getAliasPath(String applicationType, String applicationId) {
		return Constants.APP_ALIAS_PATH + Constants.SLASH + applicationType
			+ Constants.SLASH + applicationId;
	}	

	public String getCompanyPath(String applicationType, String applicationId) {
		return Constants.APP_PATH + Constants.SLASH + applicationType
				+ Constants.SLASH + applicationId;
	}
	
	public String getCompanyPathFromAlias(String applicationType, String applicationId)
		throws Exception {
		String path = getAliasPath(applicationType, applicationId);
		
		NodeOptions options = new NodeOptions();
		options.includeExtendedProperties = true;

		ClientNetContext ctx = createClient(
				path,
				ClientNetContext.GET,
				options.getOptions());
				
		NodeResponse node = ctx.readPayloadNodeResponse();
		
		return node.getMessage().getAppAliasExProperties().getReferenceUri();
	}

	public boolean doesCompanyExist(String applicationType, String applicationId)
			throws Exception {
		try {
			ClientNetContext ctx = createClient(
					getCompanyPath(applicationType, applicationId),
					ClientNetContext.GET);
			ctx.readPayloadNodeResponse();
			return true;
		} catch (SmartVaultException e) {
			if (e.getErrorCode() != ErrorCodes.OBJECT_NOT_FOUND) {
				throw e;
			}

			return false;
		}
	}
}
